import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Random;

import javax.swing.JPanel;

public class WormPanel extends JPanel implements Runnable {

	private static final int NUM_FOOD = 15;
	
	private static final int PWIDTH = 500; // size of panel
	private static final int PHEIGHT = 400;

	private Thread animator; // the thread that performs the animation
	
	private static final long NANO_TO_MILLIS = 1000000L; // multiplication-factor
	private static final long PERIOD_MILLIS = 10L; // 100FPS: each loop should take 10 millis
	private static final long PERIOD_NANOS = PERIOD_MILLIS * NANO_TO_MILLIS;
	private volatile boolean running = false; // used to stop the animation thread
	private volatile boolean isPaused = false; // used to pause game-updates
	private volatile boolean gameOver = false; // used to signal game over

	Font font = new Font("SansSerif", Font.BOLD, 14); // set up message font to display text
	FontMetrics metrics = this.getFontMetrics(font);
	private String message = "Fred"; // status string displayed on top of page

	// off screen rendering
	private Graphics dbg; 
	private Image dbImage = null;

	// FPS measuring
	private long frameCount = 0;
	private double averageFPS = 0.0;	

	private static final int NO_DELAYS_PER_YIELD = 16;
		/* Number of frames with a delay of 0 ms before the animation thread yields
		to other running threads. */
	
	private static final int MAX_FRAME_SKIPS = 5;
		/* number of frames that can be skipped in any one animation loop
		   i.e. the games state is updated but not rendered
		 */

	// UPS measuring
	private long updateGameCount = 0;
	private double averageUPS = 0.0;	

	// the worm & his food
	Worm fred;
	private int fredsDirection = Worm.EAST;
	ArrayList<Food> food = new ArrayList<Food>(NUM_FOOD);
	
	// score
	private int foodEaten;

	
	public WormPanel() {
		setBackground(Color.white);
		setPreferredSize(new Dimension(PWIDTH, PHEIGHT));
		setFocusable(true);
		requestFocus(); // focus on JPanel, so receives key events

		for (int i=0; i < NUM_FOOD; i++) { // create food on random positions
			Food f = createFood();
			food.add(f);
		}
		foodEaten = 0;
		fred = new Worm(PWIDTH,PHEIGHT); // create the worm
		
		registerKeyListeners();	// add key-listeners
	}

	
	private Food createFood() {	// create 1 food at random position
		Random r = new Random();
		int randint;
		randint = r.nextInt((PWIDTH / Worm.DOTSIZE) - 1);
		int x = randint*Worm.DOTSIZE + Worm.RADIUS;
		randint = r.nextInt((PHEIGHT / Worm.DOTSIZE) - 1);
		int y = randint*Worm.DOTSIZE + Worm.RADIUS;
		Food f = new Food(x,y);
		return f;
	}

	
	public void addNotify()
	// wait for the JPanel to be added to the JFrame before starting
	{
		super.addNotify(); // creates the peer
		startGame(); // start the thread
	}

	private void startGame()
	// initialise and start the thread 
	{
		if (animator == null || !running) {
			animator = new Thread(this);
			animator.start();
		}
	} // end of startGame()

	public void run()
	/* The frames of the animation are drawn inside the while loop. */
	{
		long beforeTime, timeDiff, sleepTime, lastFPScalc;
		beforeTime = System.nanoTime();	// start time measurement
		lastFPScalc = System.nanoTime();
		long overSleepTime = 0L; // time that has been slept to long
		long beforeSleepTime = 0L; // time measured before sleep
		int noDelays = 0;

		long excess = 0L;
		
		running = true;
		while (running) {
			gameUpdate();
			renderScreen(); // draw into buffer
			paintScreen(); // switch buffer on-screen

			// calculate sleep time
			timeDiff = System.nanoTime() - beforeTime;
			sleepTime = PERIOD_NANOS - timeDiff - overSleepTime; // loop-time
			beforeSleepTime = System.nanoTime();
			if (sleepTime > 0) {
				try {
					Thread.sleep(sleepTime / NANO_TO_MILLIS);
				} catch (InterruptedException e) { }
				overSleepTime = System.nanoTime() - beforeSleepTime - sleepTime;
			} else {	// sleepTime <= 0; frame took longer than the period
				excess -= sleepTime;	// store excess time value
				overSleepTime = 0L;
				if (++noDelays >= NO_DELAYS_PER_YIELD) {
					Thread.yield();
					noDelays = 0;
				}
			}
			beforeTime = System.nanoTime();

			// if frame animation is taking too long, update the game state without
			// rendering it, to get the updates/sec nearer to the required FPS
			int skips = 0;
			while ((excess > PERIOD_NANOS) && (skips < MAX_FRAME_SKIPS)) {
				excess -= PERIOD_NANOS;
				gameUpdate();		// update state, but do not render
				skips++;
			}
			
			// calculate average FPS and UPS every 1 second
			frameCount++;
			double diff = (System.nanoTime() - lastFPScalc) / (1000.0 * NANO_TO_MILLIS);
			if (diff > 1) {
				averageFPS = frameCount / diff;
				averageUPS = updateGameCount / diff;
				frameCount = 0;
				updateGameCount = 0;
				lastFPScalc = System.nanoTime();
			}
		}
		System.exit(0); // so window disappears
	} // end of run()

	
	private void gameUpdate() {
		if (!isPaused && !gameOver) {
			updateGameCount++;			// for stats only
			fred.move(fredsDirection);	// move fred
			
			// check if fred eats himself
			if (fred.ateMe())
				gameOver = true;
			
			// check if fred found food 
			Food toRemove = null;
			for (Food fd : food) {
				if (fred.nearHead(fd.getX(), fd.getY())) {
					toRemove = fd;
				}
			}
			
			// remove and regrow food + grow fred
			food.remove(toRemove);
			if (toRemove != null) {
				foodEaten++;
				food.add(createFood());
				fred.addHead(); // grow fred 
			}
		}
	} // end of gameUpdate()
	
	private void renderScreen() 
	// paint into buffer
	{
		if (dbImage == null) {
			dbImage = createImage(PWIDTH, PHEIGHT);
			if (dbImage == null) {
				System.out.println("dbImage is null");
				return;
			} else
				dbg = dbImage.getGraphics();
		}
	    
		// clear the background (white) and seth the font.
		dbg.setColor(Color.white);
		dbg.fillRect(0, 0, PWIDTH, PHEIGHT);
		dbg.setColor(Color.blue);
		dbg.setFont(font);
		
		// draw status message, current position and worm at current position
		dbg.drawString(message, 20, 25);
		DecimalFormat f = new DecimalFormat("#0.00"); 
		dbg.drawString("FPS: " + f.format(averageFPS),300,25);
		dbg.drawString("UPS: " + f.format(averageUPS),400,25);
		dbg.drawString("Score: " + foodEaten, 200, 375);
		
		// draw the worm && the food
		for (Food fd : food) {
			fd.draw(dbg);
		}
		fred.draw(dbg);	
		
	    if (gameOver)
	        gameOverMessage(dbg);
	} // end of renderScreen() 
	
	
	private void gameOverMessage(Graphics g) {
		String msg = "Game Over. Your score: " + foodEaten;
		int x = (PWIDTH - metrics.stringWidth(msg)) / 2;
		int y = (PHEIGHT - metrics.getHeight()) / 2;
		g.setColor(Color.red);
		g.setFont(font);
		g.drawString(msg, x, y);
	} // end of gameOverMessage()
	

	private void paintScreen()
	// use active rendering to put the buffered image on-screen
	{
		Graphics g;
		try {
			g = this.getGraphics();
			if ((g != null) && (dbImage != null))
				g.drawImage(dbImage, 0, 0, null);
			Toolkit.getDefaultToolkit().sync(); // sync the display on some
												// systems
			g.dispose();
		} catch (Exception e) {
			System.out.println("Graphics error: " + e);
		}
	} // end of paintScreen()
	
	
	private void registerKeyListeners() {
		addKeyListener(new KeyAdapter() {
			// * listen for esc, ctrl-c on the canvas to
			//   allow a convenient exit from the full screen configuration
			// * space to un/pause
			public void keyPressed(KeyEvent e) {
				int keyCode = e.getKeyCode();
				if ((keyCode == KeyEvent.VK_ESCAPE)
						|| ((keyCode == KeyEvent.VK_C) && e.isControlDown())) {
					running = false;
				} else if (keyCode == KeyEvent.VK_SPACE) {
					isPaused = !isPaused;
					if (isPaused) 
						message = "PAUSED!";
					else 
						message = "RUNNING AGAIN!";
				} else if (keyCode == KeyEvent.VK_LEFT) {
					fredsDirection = Worm.WEST;
				} else if (keyCode == KeyEvent.VK_RIGHT) {
					fredsDirection = Worm.EAST;
				} else if (keyCode == KeyEvent.VK_DOWN) {
					fredsDirection = Worm.SOUTH;
				} else if (keyCode == KeyEvent.VK_UP) {
					fredsDirection = Worm.NORTH;
				}
			}
		});
	} // end of readyForTermination()
	
	// ------------- game life cycle methods ------------
	// called by the JFrame's window listener methods

	public void resumeGame()
	// called when the JFrame is activated / deiconified
	{
		isPaused = false;
	}

	public void pauseGame()
	// called when the JFrame is deactivated / iconified
	{
		isPaused = true;
	}

	public void stopGame()
	// called when the JFrame is closing
	{
		running = false;
	}

	// ----------------------------------------------

} // end of WormPanel class
